import { isBrowser, log } from './misc/Util';
/**
 * Wrapper class for `RTCDataChannel` and `WebSocket`.
 */
var Channel = (function () {
    /**
     * Creates a channel from existing `RTCDataChannel` or `WebSocket`.
     */
    function Channel(wc, connection, options) {
        if (options === void 0) { options = { id: -1 }; }
        var _this = this;
        log.info("new connection: Me: " + wc.myId + " with " + options.id);
        this.wc = wc;
        this.connection = connection;
        this._id = options.id;
        this.rtcPeerConnection = options.rtcPeerConnection;
        this.isIntermediary = false;
        this.missedHeartbeat = 0;
        this.updateHeartbeatMsg(null);
        // Configure `send` function
        if (isBrowser) {
            connection.binaryType = 'arraybuffer';
            this.send = this.sendInBrowser;
        }
        else if (!this.rtcPeerConnection) {
            this.send = this.sendInNodeViaWebSocket;
        }
        else {
            connection.binaryType = 'arraybuffer';
            this.send = this.sendInNodeViaDataChannel;
        }
        // Configure handlers
        this.connection.onmessage = function (_a) {
            var data = _a.data;
            return wc.onMessageProxy(_this, new Uint8Array(data));
        };
        this.connection.onclose = function (evt) {
            log.info("Connection with " + _this.id + " has closed");
            _this.connection.onclose = function () { };
            _this.connection.onmessage = function () { };
            _this.connection.onerror = function () { };
            wc.topologyService.onChannelClose(evt, _this);
            if (_this.rtcPeerConnection && _this.rtcPeerConnection.signalingState !== 'closed') {
                _this.rtcPeerConnection.close();
            }
        };
        this.connection.onerror = function (evt) {
            _this.connection.onclose = function () { };
            _this.connection.onmessage = function () { };
            _this.connection.onerror = function () { };
            wc.topologyService.onChannelError(evt, _this);
            if (_this.rtcPeerConnection && _this.rtcPeerConnection.signalingState !== 'closed') {
                _this.rtcPeerConnection.close();
            }
        };
    }
    Object.defineProperty(Channel.prototype, "id", {
        get: function () {
            return this._id;
        },
        set: function (value) {
            this._id = value;
            this.updateHeartbeatMsg(this.topologyHeartbeatMsg);
        },
        enumerable: true,
        configurable: true
    });
    Channel.prototype.markAsIntermediry = function () {
        this.wc.topologyService.initIntermediary(this);
    };
    Channel.prototype.close = function () {
        if (this.connection.readyState !== 'closed' &&
            this.connection.readyState !== 'closing' &&
            this.connection.readyState !== WebSocket.CLOSED &&
            this.connection.readyState !== WebSocket.CLOSING) {
            log.info("I:" + this.wc.myId + " close connection with " + this.id);
            if (this.rtcPeerConnection && this.rtcPeerConnection.signalingState !== 'closed') {
                this.rtcPeerConnection.close();
            }
            else {
                this.connection.close();
            }
        }
    };
    Channel.prototype.closeQuietly = function () {
        this.connection.onmessage = undefined;
        this.connection.onclose = undefined;
        this.connection.onerror = undefined;
        this.close();
    };
    Channel.prototype.sendHeartbeat = function () {
        this.send(this.fullHeartbeatMsg);
    };
    Channel.prototype.updateHeartbeatMsg = function (heartbeatMsg) {
        this.topologyHeartbeatMsg = heartbeatMsg;
        this.fullHeartbeatMsg = this.wc.encode({ recipientId: this._id, content: heartbeatMsg });
    };
    Channel.prototype.sendInBrowser = function (data) {
        try {
            this.connection.send(data);
        }
        catch (err) {
            console.error('Channel send', err);
        }
    };
    Channel.prototype.sendInNodeViaWebSocket = function (data) {
        try {
            this.connection.send(data, { binary: true });
        }
        catch (err) {
            console.error('Channel send', err);
        }
    };
    Channel.prototype.sendInNodeViaDataChannel = function (data) {
        this.sendInBrowser(data.slice(0));
    };
    return Channel;
}());
export { Channel };
