import { filter, pluck } from 'rxjs/operators';
import { Subject } from 'rxjs/Subject';
import { log } from './misc/Util';
import { signaling } from './proto';
var MAXIMUM_MISSED_HEARTBEAT = 3;
var HEARTBEAT_INTERVAL = 5000;
/* WebSocket error codes */
var HEARTBEAT_ERROR_CODE = 4002;
var MESSAGE_ERROR_CODE = 4010;
var FIRST_CONNECTION_ERROR_CODE = 4011;
/* Preconstructed messages */
var heartbeatMsg = signaling.Message.encode(signaling.Message.create({ heartbeat: true })).finish();
export var SignalingState;
(function (SignalingState) {
    SignalingState[SignalingState["CONNECTING"] = 0] = "CONNECTING";
    SignalingState[SignalingState["OPEN"] = 1] = "OPEN";
    SignalingState[SignalingState["CONNECTED_WITH_FIRST_MEMBER"] = 2] = "CONNECTED_WITH_FIRST_MEMBER";
    SignalingState[SignalingState["READY_TO_JOIN_OTHERS"] = 3] = "READY_TO_JOIN_OTHERS";
    SignalingState[SignalingState["CLOSED"] = 4] = "CLOSED";
})(SignalingState || (SignalingState = {}));
/**
 * This class represents a door of the `WebChannel` for the current peer. If the door
 * is open, then clients can join the `WebChannel` through this peer. There are as
 * many doors as peers in the `WebChannel` and each of them can be closed or opened.
 */
var Signaling = (function () {
    function Signaling(wc, url) {
        // public
        this.url = url;
        this.state = SignalingState.CLOSED;
        // private
        this.wc = wc;
        this.stateSubject = new Subject();
        this.channelSubject = new Subject();
        this.rxWs = undefined;
    }
    Object.defineProperty(Signaling.prototype, "onState", {
        get: function () {
            return this.stateSubject.asObservable();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Signaling.prototype, "onChannel", {
        get: function () {
            return this.channelSubject.asObservable();
        },
        enumerable: true,
        configurable: true
    });
    /**
     * Notify Signaling server that you had joined the network and ready
     * to join new peers to the network.
     */
    Signaling.prototype.open = function () {
        if (this.state === SignalingState.CONNECTED_WITH_FIRST_MEMBER) {
            this.rxWs.send({ joined: true });
            this.setState(SignalingState.READY_TO_JOIN_OTHERS);
        }
    };
    Signaling.prototype.join = function (key) {
        var _this = this;
        if (this.state === SignalingState.READY_TO_JOIN_OTHERS) {
            throw new Error('Failed to join via signaling: connection with signaling is already opened');
        }
        if (this.state !== SignalingState.CLOSED) {
            this.close();
        }
        this.setState(SignalingState.CONNECTING);
        this.wc.webSocketBuilder.connect(this.getFullURL(key))
            .then(function (ws) {
            _this.setState(SignalingState.OPEN);
            _this.rxWs = _this.createRxWs(ws);
            _this.rxWs.onMessage.subscribe(function (msg) {
                switch (msg.type) {
                    case 'heartbeat':
                        _this.missedHeartbeat = 0;
                        break;
                    case 'isFirst':
                        if (msg.isFirst) {
                            _this.setState(SignalingState.READY_TO_JOIN_OTHERS);
                        }
                        else {
                            _this.wc.webRTCBuilder.connectOverSignaling({
                                onMessage: _this.rxWs.onMessage.pipe(filter(function (_a) {
                                    var type = _a.type;
                                    return type === 'content';
                                }), pluck('content')),
                                send: function (m) { return _this.rxWs.send({ content: m }); },
                            })
                                .then(function (ch) {
                                _this.setState(SignalingState.CONNECTED_WITH_FIRST_MEMBER);
                                ch.markAsIntermediry();
                            })
                                .catch(function (err) {
                                _this.rxWs.close(FIRST_CONNECTION_ERROR_CODE, "Failed to join over Signaling: " + err.message);
                            });
                        }
                        break;
                }
            });
            _this.startHeartbeat();
        })
            .catch(function (err) { return _this.setState(SignalingState.CLOSED); });
    };
    /**
     * Close the `WebSocket` with Signaling server.
     */
    Signaling.prototype.close = function () {
        if (this.rxWs) {
            this.rxWs.close(1000);
        }
    };
    Signaling.prototype.setState = function (state) {
        var _this = this;
        if (this.state !== state) {
            this.state = state;
            this.stateSubject.next(state);
            if (state === SignalingState.READY_TO_JOIN_OTHERS) {
                this.wc.webRTCBuilder.onChannelFromSignaling({
                    onMessage: this.rxWs.onMessage.pipe(filter(function (_a) {
                        var type = _a.type;
                        return type === 'content';
                    }), pluck('content')),
                    send: function (msg) { return _this.rxWs.send({ content: msg }); },
                }).subscribe(function (ch) { return _this.channelSubject.next(ch); });
            }
        }
    };
    Signaling.prototype.startHeartbeat = function () {
        var _this = this;
        this.missedHeartbeat = 0;
        this.heartbeatInterval = global.setInterval(function () {
            try {
                _this.missedHeartbeat++;
                if (_this.missedHeartbeat >= MAXIMUM_MISSED_HEARTBEAT) {
                    throw new Error('Too many missed heartbeats');
                }
                _this.rxWs.heartbeat();
            }
            catch (err) {
                global.clearInterval(_this.heartbeatInterval);
                log.info('Closing connection with Signaling. Reason: ' + err.message);
                _this.rxWs.close(HEARTBEAT_ERROR_CODE, 'Signaling is not responding');
            }
        }, HEARTBEAT_INTERVAL);
    };
    Signaling.prototype.createRxWs = function (ws) {
        var _this = this;
        var subject = new Subject();
        ws.binaryType = 'arraybuffer';
        ws.onmessage = function (evt) {
            try {
                subject.next(signaling.Message.decode(new Uint8Array(evt.data)));
            }
            catch (err) {
                ws.close(MESSAGE_ERROR_CODE, err.message);
            }
        };
        ws.onerror = function (err) {
            log.debug('Signaling ERROR', err);
            subject.error(err);
        };
        ws.onclose = function (closeEvt) {
            clearInterval(_this.heartbeatInterval);
            _this.setState(SignalingState.CLOSED);
            subject.complete();
            log.info("Connection with Signaling '" + _this.url + "' closed: " + closeEvt.code + ": " + closeEvt.reason);
        };
        return {
            onMessage: subject.asObservable(),
            send: function (msg) {
                if (ws.readyState !== WebSocket.CLOSING && ws.readyState !== WebSocket.CLOSED) {
                    ws.send(signaling.Message.encode(signaling.Message.create(msg)).finish());
                }
            },
            heartbeat: function () {
                if (ws.readyState !== WebSocket.CLOSING && ws.readyState !== WebSocket.CLOSED) {
                    ws.send(heartbeatMsg);
                }
            },
            close: function (code, reason) {
                if (code === void 0) { code = 1000; }
                if (reason === void 0) { reason = ''; }
                return ws.close(code, reason);
            },
        };
    };
    Signaling.prototype.getFullURL = function (params) {
        if (this.url.endsWith('/')) {
            return this.url + params;
        }
        else {
            return this.url + '/' + params;
        }
    };
    return Signaling;
}());
export { Signaling };
